from multiprocessing import cpu_count
import socket


def get_process_count(starts):
    """ 
    Using the number of starts per graph and machine info, calculate
    the number of processes to spawn 
    
    :param starts:
    :type starts:
    
    :returns: 
    :rtype: 
    """
    if socket.gethostname().split('.')[0] == 'leibnitz':
        num_processes = 30
        num_processes = max((1, num_processes / starts))
    elif socket.gethostname().split('.')[0] == 'mars':
        num_processes = 21
        num_processes = max((1, num_processes / starts))
    elif socket.gethostname().split('.')[0] == 'saturn':
        num_processes = 12
        num_processes = max((1, num_processes / starts))
    elif socket.gethostname().split('.')[0] == 'hooke':
        num_processes = 22
        num_processes = max((1, num_processes / starts))
    else:
        # Setting the number  of parallel running processes  to the number
        # of cores minus 7% for breathing room
        num_processes = cpu_count() - int(0.07 * cpu_count())
        num_processes = max((1, num_processes / starts))
        
    return num_processes



""" This module contains clingo interaction functions """
from __future__ import print_function
from gunfolds.utils.clingo import clingo
from gunfolds.conversions import g2clingo, g2wclingo, msl_jclingo2g
import time


uprogram = """
{ edge1(X,Y) } :- node(X),node(Y).
edge(X,Y,1) :- edge1(X,Y).
edge(X,Y,L) :- edge(X,Z,L-1),edge1(Z,Y), L <= U, u(U).
derived_edgeu(X,Y) :- edge(X,Y,L), u(L).
derived_confu(X,Y) :- edge(Z,X,L), edge(Z,Y,L),node(X),node(Y),node(Z),X < Y, L < U, u(U).
:- edgeu(X,Y), not derived_edgeu(X,Y),node(X),node(Y).
:- not edgeu(X,Y), derived_edgeu(X,Y),node(X),node(Y).
:- derived_confu(X,Y), not confu(X,Y),node(X),node(Y), X < Y.
:- not derived_confu(X,Y), confu(X,Y),node(X),node(Y), X < Y.
    """
wuprogram = """
    { edge1(X,Y) } :- node(X), node(Y).
    path(X,Y,1) :- edge1(X,Y).
    path(X,Y,L) :- path(X,Z,L-1), edge1(Z,Y), L <= U, u(U).
    edgeu(X,Y) :- path(X,Y,L), u(L).
    confu(X,Y) :- path(Z,X,L), path(Z,Y,L), node(X;Y;Z), X < Y, L < U, u(U).
    :~ edgeh(X,Y,W), not edgeu(X,Y). [W,X,Y,1]
    :~ no_edgeh(X,Y,W), edgeu(X,Y). [W,X,Y,1]
    :~ confh(X,Y,W), not confu(X,Y). [W,X,Y,2]
    :~ no_confh(X,Y,W), confu(X,Y). [W,X,Y,2]
    """


def rate(u):
    """
    :param u: maximum under sampling rate 
    :type u: integer
    
    :returns: predicate for under sampling rate
    :rtype: string
    """
    s = "u("+str(u)+")."
    return s


def g2clingo_msl(g):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    return g2clingo(g, directed='edgeu', bidirected='confu', both_bidirected=True)


def msl_command(g, urate=2, exact=True):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
  
    :param urate: maximum undersampling rate to consider
    :type urate: integer
    
    :param exact: If true, run clingo in exact mode. If false, run clingo in optimization mode 
    :type exact: boolean
    
    :returns: 
    :rtype: 
    """
    if exact:
        command = rate(urate) + ' ' + g2clingo_msl(g) + ' ' + uprogram
    else:
        command = rate(urate) + ' ' + g2wclingo(g) + ' ' + wuprogram
    command = command.encode().replace(b"\n", b" ")
    return command


def msl(g, capsize, exact=True, configuration="tweety", urate=2, timeout=0, pnum=None):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param capsize: maximum number of candidates to return
    :type capsize: integer
    
    :param exact: If true, run clingo in exact mode. If false, run clingo in optimization mode 
    :type exact: boolean
    
    :param urate: maximum undersampling rate to consider
    :type urate: integer
    
    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer

    :param configuration: Select configuration based on problem type

        - ``frumpy`` : Use conservative defaults
        - ``jumpy`` : Use aggressive defaults
        - ``tweety`` : Use defaults geared towards asp problems
        - ``handy`` : Use defaults geared towards large problems
        - ``crafty`` : Use defaults geared towards crafted problems
        - ``trendy`` : Use defaults geared towards industrial problems
    :type configuration: string
    
    :param pnum: number of parallel threads to run ``clingo`` on
    :type pnum: integer
    
    :returns: 
    :rtype: 
    """
    return clingo(msl_command(g, urate=urate, exact=True),
                  capsize=capsize, convert=msl_jclingo2g, timeout=timeout, configuration=configuration, pnum=pnum)


def rasl_msl(g, capsize, urate=2, timeout=0, pnum=None):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param capsize: maximum number of candidates to return
    :type capsize: integer
    
    :param urate: maximum undersampling rate to consider
    :type urate: integer
    
    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer
    
    :param pnum: number of parallel threads to run ``clingo`` on
    :type pnum: integer
    
    :returns: 
    :rtype: 
    """
    r = set()
    remain_time = timeout
    for i in range(2, urate + 1):
        startTime = int(round(time.time()))
        k = msl(g, capsize, urate=i, pnum=pnum, timeout=remain_time, exact=True)
        endTime = int(round(time.time()))
        remain_time = max(0, remain_time - (endTime - startTime))
        r = r | {(x[0], i) for x in k}
        capsize = max(0, capsize - len(r))
        if not (capsize and remain_time):
            break
    return r



import sys
import threading


class KThread(threading.Thread):
    """A subclass of threading.Thread, with a kill() method.

    Come from:
    Kill a thread in Python:
    http://mail.python.org/pipermail/python-list/2004-May/260937.html
    """

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.killed = False
        self.__run_backup = None

    def start(self):
        """Start the thread."""
        self.__run_backup = self.run
        self.run = self.__run      # Force the Thread to install our trace.
        threading.Thread.start(self)

    def __run(self):
        """Hacked run function, which installs the trace."""
        sys.settrace(self.globaltrace)
        self.__run_backup()
        self.run = self.__run_backup

    def globaltrace(self, why):
        if why == 'call':
            return self.localtrace
        return None

    def localtrace(self, why):
        if self.killed:
            if why == 'line':
                raise SystemExit()
        return self.localtrace

    def kill(self):
        self.killed = True


class TIMEOUT_EXCEPTION(Exception):
    """function run timeout"""
    pass


def timeout(seconds):
    def timeout_decorator(func):

        def _new_func(oldfunc, result, oldfunc_args, oldfunc_kwargs):
            result.append(oldfunc(*oldfunc_args, **oldfunc_kwargs))

        def _(*args, **kwargs):
            result = []
            new_kwargs = {
                # create new args for _new_func, because we want to get the
                # func return val to result list
                'oldfunc': func,
                'result': result,
                'oldfunc_args': args,
                'oldfunc_kwargs': kwargs
            }

            thd = KThread(target=_new_func, args=(), kwargs=new_kwargs)
            thd.start()
            thd.join(seconds)
            alive = thd.isAlive()
            thd.kill()  # kill the child thread

            if alive:
                raise TIMEOUT_EXCEPTION(
                    'function run too long, timeout %d seconds.' % seconds)
            else:
                if result:
                    return result[0]
                return result

        _.__name__ = func.__name__
        _.__doc__ = func.__doc__
        return _

    return timeout_decorator



# tools to construct (random) graphs
from gunfolds.conversions import nx2graph, nxbp2graph, graph2nx
from gunfolds.utils import ecj
from itertools import combinations
import networkx as nx
from networkx.utils.random_sequence import powerlaw_sequence
import numpy as np
import random


def edgelist(g):  # directed
    """
    Returns a list of tuples for edges of ``g``
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: a list of tuples for edges of ``g``
    :rtype: list of tuples
    """
    l = []
    for n in g:
        l.extend([(n, e) for e in g[n] if g[n][e] in (1, 3)])
    return l


def inedgelist(g):  # missing directed iterator
    """
    Iterate over the list of tuples for edges of ``g``
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    """
    n = len(g)
    for v in g:
        for i in range(1, n + 1):
            if not (i in g[v]):
                yield v, i
            elif g[v][i] not in (1, 3):
                yield v, i


def inbedgelist(g):  # missing bidirected iterator
    """
    Iterate over the list of tuples for edges of ``g``
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    """
    for v in g:
        for w in g:
            if v != w:
                if not (w in g[v]):
                    yield v, w
                elif g[v][w] not in (2, 3):
                    yield v, w


def bedgelist(g):
    """ 
    Bidirected edge list with flips 
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: a list of tuples for bidirected edges of ``g``
    :rtype: list of tuples
    """
    l = []
    for n in g:
        l.extend([tuple(sorted((n, e))) for e in g[n] if g[n][e] in (2, 3)])
    l = list(set(l))
    l = l + list(map(lambda x: (x[1], x[0]), l))
    return l


def undedgelist(g, exclude_bi=False):
    """
    Returns a list of tuples for undirected edges of ``g`` with nodes
    sorted in ascending order.
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param exclude_bi: if True, only converts directed edges to undirected
        edges. If False, converts directed and bidirected edges to undirected edges
    :type exclude_bi: boolean
    
    :returns: a list of tuples for undirected edges of ``g`` with nodes
              sorted in ascending order
    :rtype: list
    """
    l = []
    if exclude_bi:
        for n in g:
            l.extend([(n, e) if n <= e else (e, n) for e in g[n] if g[n][e] in (1, 3)])
    else:
        for n in g:
            l.extend([(n, e) if n <= e else (e, n) for e in g[n] if g[n][e] in (1, 2, 3)])
    return set(l)


def superclique(n):
    """ 
    Returns a Graph with all possible edges
    
    :param n: number of nodes
    :type n: integer
    
    :returns: a Graph with all possible edges
    :rtype: dictionary (``gunfolds`` graph)
    """
    g = {}
    for i in range(n):
        g[i + 1] = {j + 1: 3 for j in range(n) if j != i}
        g[i + 1][i + 1] = 1
    return g


def fullyconnected(n):
    """
    Returns a graph with all possible directed edges

    :param n: number of nodes
    :type n: integer

    :returns: a graph with all possible directed edges
    :rtype: dictionary(``gunfolds`` graph)
    """
    g = {}
    for i in range(n):
        g[i + 1] = {j + 1: 1 for j in range(n)}
    return g


def complement(G):
    """ 
    Returns the complement of ``G``
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :returns: the complement of ``G``
    :rtype: dictionary (``gunfolds`` graph)
    """
    n = len(G)
    sq = superclique(n)
    for v in G:
        for w in G[v]:
            sq[v][w] = sq[v][w] - G[v][w]
            if sq[v][w] == 0:
                del sq[v][w]
    return sq


def gtranspose(G):
    """ 
    Transpose (rev. edges of) ``G``
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :returns: Transpose of given graph
    :rtype: dictionary (``gunfolds`` graph)
    """
    GT = {u: {} for u in G}
    for u in G:
        for v in G[u]:
            if G[u][v] in (1, 3):
                GT[v][u] = 1        # Add all reverse edges
    return GT


def scale_free(n, alpha=0.7, beta=0.25, delta_in=0.2, delta_out=0.2):
    """
    (Copied from scale_free function in networkx should i need to add any citation)

    :param n: number of nodes
    :type n: integer
    
    :param alpha: probability for adding a new node connected to an existing node
    :type alpha: float
    
    :param beta: probability for adding an edge between two existing nodes.
    :type beta: float
    
    :param delta_in: bias for choosing nodes from in-degree
    :type delta_in: float
    
    :param delta_out: bias for choosing nodes from out-degree distribution.
    :type delta_out: float
    
    :returns: 
    :rtype: 
    """
    g = nx.scale_free_graph(n, alpha=alpha,
                            beta=beta,
                            delta_in=delta_in, delta_out=delta_out)
    g = nx2graph(g)
    g = gtranspose(g)
    addAring(g)
    return g


def randH(n, d1, d2):
    """ 
    Generate a random H with ``n`` nodes 
    
    :param n: number of nodes
    :type n: integer
    
    :param d1: number of additional edges to the ring
    :type d1: integer
    
    :param d2: (ask) number of random permutations
    :type d2: integer
    
    :returns: a random graph with ``n`` nodes 
    :rtype: dictionary (``gunfolds`` graph)
    """
    g = ringmore(n, d1)
    pairs = [x for x in combinations(g.keys(), 2)]
    for p in np.random.permutation(pairs)[:d2]:
        g[p[0]][p[1]] = g[p[0]].get(p[1], 0) + 2
        g[p[1]][p[0]] = g[p[1]].get(p[0], 0) + 2
    return g


def ring(n):
    """
    Create a Ring Graph with ``n`` number of nodes

    :param n: number of nodes
    :type n: integer
    
    :returns: a Ring Graph with ``n`` number of nodes
    :rtype: dictionary (``gunfolds`` graph)
    """
    g = {}
    for i in range(1, n):
        g[i] = {i + 1: 1}
    g[n] = {1: 1}
    return g


def addAring(g):
    """
    Add a ring to ``g`` in place
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    """
    for i in range(1, len(g)):
        if g[i].get(i + 1) == 2:
            g[i][i + 1] = 3
        else:
            g[i][i + 1] = 1
    if g[i].get(1) == 2:
        g[i][1] = 3
    else:
        g[i][1] = 1


def upairs(n, k):
    """
    Returns ``n`` unique nonsequential pairs
    
    :param n: (ask)
    :type n: integer
    
    :param k: (ask)
    :type k: integer
    
    :returns: n unique nonsequential pairs
    :rtype: list
    """
    s = set()
    for p in np.random.randint(n, size=(3 * k, 2)):
        if p[1] - p[0] == 1:
            continue
        s.add(tuple(p))
    return list(s)[:k]


def ringarcs(g, n):
    """
    (Ask)

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param n: (ask)
    :type n: integer
    
    :returns: 
    :rtype: dictionary (``gunfolds`` graph)
    """
    for edge in upairs(len(g), n):
        g[edge[0] + 1][edge[1] + 1] = 1
    return g


def ringmore(n, m):
    """
    Returns a ``n`` node ring graph with ``m`` additional edges

    :param n: number of nodes
    :type n: integer
    
    :param m: number of additional edges
    :type m: integer
    
    :returns: a ``n`` node ring graph with ``m`` additional edges
    :rtype: dictionary (``gunfolds`` graph)
    """
    return ringarcs(ring(n), m)


def digonly(H):
    """   
    Returns a subgraph of ``H`` contatining all directed edges of ``H``

    :param H: ``gunfolds`` graph
    :type H: dictionary (``gunfolds`` graph)
    
    :returns: a subgraph of ``H`` contatining all directed edges of ``H``
    :rtype: dictionary (``gunfolds`` graph)
    """

    g = {n: {} for n in H}
    for v in g:
        g[v] = {w: 1 for w in H[v] if not H[v][w] == 2}
    return g


def _OCE(g1, g2):
    """
    Omission/commision error of ``g1`` referenced to ``g2``
    
    :param g1: the graph to check
    :type g1: dictionary (``gunfolds`` graph)
    
    :param g2: the ground truth graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: Omission/commision error for directed and bidirected edges
    :rtype: dictionary
    """
    s1 = set(edgelist(g1))
    s2 = set(edgelist(g2))
    omitted = len(s2 - s1)
    comitted = len(s1 - s2)

    s1 = set(bedgelist(g1))
    s2 = set(bedgelist(g2))
    bomitted = len(s2 - s1)//2
    bcomitted = len(s1 - s2)//2

    return {'directed': (omitted, comitted),
            'bidirected': (bomitted, bcomitted),
            'total': (omitted+bomitted, comitted+bcomitted)}


def _normed_OCE(g1, g2):
    """
    Return omission and comission errors for directed and
    bidirected edges.

    Omission error is normalized by the number of edges present
    in the ground truth. Commision error is normalized by the
    number of possible edges minus the number of edges present
    in the ground truth.
    
    :param g1: the graph to check
    :type g1: dictionary (``gunfolds`` graph)
    
    :param g2: the ground truth graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: normalized omission and comission errors for directed and
              bidirected edges.
    :rtype: dictionary
    """
    def sdiv(x, y):
        if y < 1.:
            return 0.
        return x/y

    n = len(g2)
    gt_DEN = float(len(edgelist(g2)))  # number of d  edges in GT
    gt_BEN = 0.5 * len(bedgelist(g2))  # number of bi edges in GT
    DEN = n*n                          # all posible directed edges
    BEN = n*(n-1)/2                    # all possible bidirected edges
    err = OCE(g1, g2)
    nerr = {'directed': (sdiv(err['directed'][0], gt_DEN),
                         sdiv(err['directed'][1], (DEN - gt_DEN))),
            'bidirected': (sdiv(err['bidirected'][0], gt_BEN),
                           sdiv(err['bidirected'][1], (BEN - gt_BEN))),
            'total': (sdiv((err['directed'][0]+err['bidirected'][0]), (gt_DEN+gt_BEN)),
                      sdiv((err['directed'][1]+err['bidirected'][1]),
                           (DEN+BEN - gt_BEN - gt_DEN)))
            }
    return nerr


def _undirected_OCE(g1, g2):
    """
    Returns omission/commision error of ``g1`` referenced to ``g2``
    if both are undirected graphs.
    
    :param g1: the graph to check
    :type g1: dictionary (``gunfolds`` graph)
    
    :param g2: the ground truth graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: omission and comission errors for undirected edges
    :rtype: dictionary
    """
    
    s1 = set(undedgelist(g1))
    s2 = set(undedgelist(g2))
    omitted = len(s2 - s1)
    comitted = len(s1 - s2)
    return {'undirected': (omitted, comitted)}


def _normed_undirected_OCE(g1, g2):
    """
    Return omission and comission errors for undirected edges.

    Omission error is normalized by the number of edges present
    in the ground truth. Commision error is normalized by the
    number of possible edges minus the number of edges present
    in the ground truth.

    :param g1: the graph to check
    :type g1: dictionary (``gunfolds`` graph)
    
    :param g2: the ground truth graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: omission and comission errors for normalized undirected edges
    :rtype: dictionary
    """

    def sdiv(x, y):
        if y < 1.:
            return 0.
        return x / y

    n = len(g2)
    gt_DEN = float(len(undedgelist(g2)))  # number of undirected  edges in GT
    DEN = n * n + n  # all posible undirected edges
    err = _undirected_OCE(g1, g2)
    nerr = {'undirected': (err['undirected'][0]/gt_DEN,
            sdiv(err['undirected'][1], (DEN - gt_DEN)))}
    return nerr


def OCE(g1, g2, normalized=False, undirected=False):
    """
    Return omission and comission errors for graphs.

    :param g1: the graph to check
    :type g1: dictionary (``gunfolds`` graph)
    
    :param g2: the ground truth graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :param normalized: If True, returns normalized error and vice versa
    :type normalized: boolean
    
    :param undirected: If True, returns undirected error and vice versa
    :type undirected: boolean
    
    :returns: omission and comission errors for graphs
    :rtype: dictionary
    """
    if normalized:
        if undirected:
            err = _normed_undirected_OCE(g1, g2)
        else:
            err = _normed_OCE(g1, g2)
    else:
        if undirected:
            err = _undirected_OCE(g1, g2)
        else:
            err = _OCE(g1, g2)
    return err


def clean_leaf_nodes(g):
    """
    Removes leaf_nodes of the given graph ``g``

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    """
    for v in g:
        g[v] = {w: g[v][w] for w in g[v] if g[v][w] > 0}


def cerror(d):
    """
    Returns normalized comission error

    :param d: calculated error
    :type d: dictionary
    
    :returns: normalized comission error
    :rtype: float
    """
    return d['OCE']['directed'][1] / np.double(len(d['gt']['graph']) ** 2 - len(edgelist(d['gt']['graph'])))


def oerror(d):
    """
    Returns normalized omission error

    :param d: calculated error
    :type d: dictionary
    
    :returns: normalized omission error
    :rtype: float
    """
    return d['OCE']['directed'][0] / np.double(len(edgelist(d['gt']['graph'])))


def bidirected_no_fork(g):
    """
    (Ask)

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: 
    :rtype: boolean
    """
    be = bedgelist(g)
    T = gtranspose(g)
    for e in be:
        if not set(T[e[0]].keys()) & set(T[e[1]].keys()):
            return True
    return False


def no_parents(g):
    """
    Checks if there exists a node that has no parents in the given graph ``g``.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: True, if there exists a node with no parents. False, otherwise.
    :rtype: boolean
    """
    T = gtranspose(g)
    for n in T:
        if not T[n]:
            return True
    return False


def no_children(g):
    """
    Checks if there exists a node that has no children in the given graph ``g``.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: True, if there exists a node with no children. False, otherwise.
    :rtype: boolean
    """
    for n in g:
        if not g[n]:
            return True
    return False


def scc_unreachable(g):
    """
    Checks if there exists a strongly connected component in the given graph ``g`` that is unreachable.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: True, if there exists SCC that is unreachable. False, otherwise.
    :rtype: boolean
    """
    if bidirected_no_fork(g):
        return True
    if no_parents(g):
        return True
    if no_children(g):
        return True
    return False

# unlike functions from traversal package these do no checking


def addanedge(g, e):
    """
    Adds an edge ``e`` from the given graph ``g``

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param e: edge to be added
    :type e: pair of integers
    """
    g[e[0]][e[1]] = 1


def delanedge(g, e):
    """
    Deletes an edge ``e`` from the given graph ``g``

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param e: edge to be deleted
    :type e: pair of integers
    """
    g[e[0]].pop(e[1], None)


def addedges(g, es):
    """
    Adds the edges in the list ``es`` for given graph ``g``.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param es: list of edges to be added
    :type es: list of pairs of integers
    """
    for e in es:
        addanedge(g, e)


def deledges(g, es):
    """
    Deletes the edges in the list ``es`` for given graph ``g``.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param es: list of edges to be deleted
    :type es: list of pairs of integers
    """
    for e in es:
        delanedge(g, e)


def isdedgesubset(g2star, g2):
    """
    Checks if ``g2star`` directed edges are a subset of those of ``g2``
    
    :param g2star: ``gunfolds`` graph to be checked
    :type g2star: dictionary (``gunfolds`` graph)
    
    :param g2:  ``gunfolds`` graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: True, if ``g2star`` directed edges are a subset of those of ``g2`` and vice versa
    :rtype: boolean
    """
    for n in g2star:
        for h in g2star[n]:
            if h in g2[n]:
                # if g2star has a directed edge and g2 does not
                if g2star[n][h] in (1, 3) and g2[n][h] == 2:
                    return False
            else:
                return False
    return True


def isedgesubset(g2star, g2):
    """
    Checks if all ``g2star`` edges are a subset of those of ``g2``
    
    :param g2star: ``gunfolds`` graph to be checked
    :type g2star: dictionary (``gunfolds`` graph)
    
    :param g2:  ``gunfolds`` graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: True, if all ``g2star`` edges are a subset of those of ``g2`` and vice versa
    :rtype: boolean
    """
    for n in g2star:
        for h in g2star[n]:
            if h in g2[n]:
                # Everything is a subset of 3 (both edge types)
                if g2[n][h] != 3:
                    # Either they both should have a directed edge, or
                    # both should have a bidirected edge
                    if g2star[n][h] != g2[n][h]:
                        return False
            else:
                return False
    return True


def degree_ring(n, d):
    """
    Generate a ring graph with ``n`` nodes and average degree ``d``
    
    :param n: number of nodes
    :type n: integer
    
    :param d: degree
    :type d: integer
    
    :returns: a ring graph with ``n`` nodes and average degree ``d``
    :rtype: dictionary(``gunfolds`` graph)
    """
    g = nx.expected_degree_graph([d-1]*n)
    gn = nx2graph(g)
    addAring(gn)
    return gn


def density(g):
    """
    Returns the density of the directed part of the graph ``g``.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: the density of the directed part of the graph
    :rtype: float
    """
    return len(edgelist(g)) / np.double(len(g) ** 2)


def udensity(g):
    """
    Returns the density of a undersampled graph that simultaneously accounts for directed and bidirected edges of ``g``.

    :param g: `gunfolds` graph
    :type g: dictionary (`gunfolds` graph)
    
    :returns: the density of the given undersampled graph
    :rtype: float
    """
    return (len(edgelist(g))+len(bedgelist(g))/2) / np.double(len(g)**2 + len(g)*(len(g)-1)/2)


def mean_degree_graph(node_num, degree):
    """
    Generates a random graph with ``node_num``, nodes and mean outgoing degree of ``degree``.

    :param node_num: number of nodes
    :type node_num: integer
    
    :param degree: degree
    :type degree: integer
    
    :returns: a random graph with mean outgoing degree of the given graph
    :rtype: dictionary(``gunfolds`` graph)
    """
    g = nx.fast_gnp_random_graph(node_num, degree/node_num, directed=True)
    g = nx2graph(g)
    return g


def pow_degree_graph(node_num, degree):
    """
    Generates a graph by powerlaw sequence of ``degree`` constructed using the Havel-Hakimi algorithm.

    :param node_num: number of nodes
    :type node_num: integer
    
    :param degree: degree
    :type degree: integer
    
    :returns: a graph constructed using the Havel-Hakimi algorithm.
    :rtype: dictionary(``gunfolds`` graph)
    """
    while True:
        try:
            sequence = powerlaw_sequence(node_num, exponent=degree)
            g = nx.havel_hakimi_graph([int(x) for x in sequence])
            break
        except nx.NetworkXError:
            continue
    g = nx2graph(g)
    return g


def bp_mean_degree_graph(node_num, degree, seed=None):
    """
    Generates a random bipartite graph with ``node_num``, nodes and mean outgoing degree (``degree``)

    :param node_num: number of nodes
    :type node_num: integer
    
    :param degree: degree
    :type degree: float
    
    :param seed: random seed
    :type seed: integer
    
    :returns: a random bipartite graph with ``node_num``, nodes and mean outgoing degree of ``degree``
    :rtype: dictionary(``gunfolds`` graph)
    """
    G = nx.bipartite.random_graph(node_num, node_num, degree/node_num, seed=seed)
    g = nxbp2graph(G)
    return g


# this function does not work yet - WIP
def bp_pow_degree_graph(node_num, degree, prob=0.7):
    """
    Generates a bipartite graph by powerlaw sequence of ``degree`` constructed using the Havel-Hakimi algorithm.

    :param node_num: number of nodes
    :type node_num: integer
    
    :param degree: degree
    :type degree: integer
    
    :param prob: probability 
    :type prob: float
    
    :returns: a bipartite graph constructed using the Havel-Hakimi algorithm.
    :rtype: dictionary(``gunfolds`` graph)
    """
    while True:
        try:
            sequence = powerlaw_sequence(node_num, exponent=degree)
            g = nx.bipartite.preferential_attachment_graph([int(x) for x in sequence], prob)
            break
        except nx.NetworkXError:
            continue
    g = nxbp2graph(g)
    return g


def remove_tril_singletons(T):
    """
    Ensure that the DAG resulting from this matrix will not have
    singleton nodes not connected to anything.

    :param T: lower triangular matrix representing a DAG
    :type T: numpy array
    
    :returns: adjacency matrix where no singleton nodes are connected to anything
    :rtype: numpy array 
    """
    N = T.shape[0]
    neighbors = T.sum(0) + T.sum(1)
    idx = np.where(neighbors == 0)
    if idx:
        for i in idx[0]:
            v1 = i
            while i == v1:
                v1 = np.random.randint(0, N-1)
            if i > v1:
                T[i][v1] = 1
            elif i < v1:
                T[v1][i] = 1
    return T


def randomTRIL(N, degree=5, connected=False):
    """
    Generate a random triangular matrix

    https://stackoverflow.com/a/56514463
    
    :param N: size of the matrix
    :type N: integer
    
    :param degree: degree
    :type degree: integer
    
    :param connected: (Ask)
    :type connected: boolean
    
    :returns: a random triangular adjacency matrix
    :rtype: numpy array 
    """
    mat = [[0 for x in range(N)] for y in range(N)]
    for _ in range(N):
        for j in range(degree):
            v1 = np.random.randint(0, N-1)
            v2 = np.random.randint(0, N-1)
            if v1 > v2:
                mat[v1][v2] = 1
            elif v1 < v2:
                mat[v2][v1] = 1
    mat = np.asarray(mat, dtype=np.uint8)
    if connected:
        mat = remove_tril_singletons(mat)
    return mat


def randomDAG(N, degree=5, connected=True):
    """
    Generates a random DAG

    :param N: number of nodes
    :type N: integer
    
    :param degree: degree
    :type degree: integer
    
    :param connected: If true, returns a connected DAG
    :type connected: boolean
    
    :returns: a random DAG
    :rtype: NetworkX graph
    """
    adjacency_matrix = randomTRIL(N, degree=degree,
                                  connected=connected)
    rows, cols = np.where(adjacency_matrix == 1)
    edges = zip(rows.tolist(), cols.tolist())
    gr = nx.DiGraph()
    gr.add_edges_from(edges)
    if connected:
        components = [x for x in nx.algorithms.components.weakly_connected_components(gr)]
        if len(components) > 1:
            for component in components[1:]:
                v1 = random.choice(tuple(components[0]))
                v2 = random.choice(tuple(component))
                gr.add_edge(v1, v2)
    assert nx.is_directed_acyclic_graph(gr)
    assert nx.algorithms.components.is_weakly_connected(gr)
    return gr


def shift_labels(g, shift):
    """
    Returns same graph with shifted labels

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param shift: number of vertices to be shifted
    :type shift: integer
    
    :returns: same graph with shifted labels
    :rtype: dictionary(``gunfolds`` graph)
    """
    new_g = {}
    for v in g:
        new_g[v+shift] = {}
        for w in g[v]:
            new_g[v+shift][w+shift] = g[v][w]
    return new_g


def shift_list_labels(glist):
    """
    Shifts the labels for a list of graphs

    :param glist: a list of graphs that are undersampled versions of
        the same system
    :type glist: list of dictionaries (``gunfolds`` graphs)
    
    :returns: shifted list of graphs
    :rtype: list
    """
    if len(glist) < 2:
        return glist
    components = [glist[0]]
    shift = len(glist[0])
    for i in range(1, len(glist)):
        components.append(shift_labels(glist[i], shift))
        shift += len(glist[i])
    return components


def merge_list(glist):
    """
    Merge the graphs in the list

    :param glist: a list of graphs that are undersampled versions of
        the same system
    :type glist: list of dictionaries (``gunfolds`` graphs)
    
    :returns: merged graph of a list
    :rtype: dictionary(``gunfolds`` graph)
    """
    g = {}
    for e in glist:
        g.update(e)
    return g


def subgraph(g, nodes):
    """
    Returns a subgraph of ``g`` that consists of ``nodes`` and their
    interconnections.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)

    :param nodes: integer valued nodes to include
    :type nodes: list
    
    :returns: a subgraph of ``g`` that consists of ``nodes`` and their
              interconnections.
    :rtype: dictionary(``gunfolds`` graph)
    """
    nodes = set(nodes)
    sg = {}
    for node in nodes:
        sg[node] = {x: g[node][x] for x in g[node]}
    return sg


def gcd4scc(SCC):
    # first check if there is at least a single simple loop
    """
    Returns the greatest common divisor of the strongly connected component

    :param SCC: a strongly connected component
    :type SCC: dictionary (``gunfolds`` graph)
    
    :returns: the greatest common divisor of the strongly connected component
    :rtype: integer
    """
    x = np.sum([selfloop(_, SCC) for _ in SCC])
    if x > 0:
        return 1
    g = graph2nx(SCC)
    return ecj.listgcd([len(x) for x in nx.simple_cycles(g)])


def ensure_gcd1(scc):
    """
    If ``scc``'s loop structure is not ``gcd=1`` pick a random node and
    add a self-loop

    :param scc: a strongly connected component
    :type scc: dictionary (``gunfolds`` graph)
    
    :returns: a graph with ``gcd=1``
    :rtype: dictionary (``gunfolds`` graph)
    """
    if gcd4scc(scc) > 1:
        a = random.choice([*scc])
        scc[a][a] = 1
    return scc


def update_graph(g, g2):
    """
    Update ``g`` with connections (or nodes) from ``g2``. Both must be at ``u=1``
    
    :param g: graph to be updated
    :type g: dictionary (``gunfolds`` graph)
    
    :param g2: reference graph
    :type g2: dictionary (``gunfolds`` graph)
    
    :returns: updated ``g`` with respect to ``g2``
    :rtype: dictionary (``gunfolds`` graph)
    """
    for v in g2:
        if v in g:
            g[v].update(g2[v])
        else:
            g[v] = g2[v].copy()
    return g


def merge_graphs(glist):
    """
    Merge a list of graphs at ``u=1`` into a single new graph ``g``
    
    :param glist: a list of graphs that are undersampled versions of
        the same system
    :type glist: list of dictionaries (``gunfolds`` graphs)
    
    :returns: a new  graph by merging a list of graphs at ``u=1``
    :rtype: dictionary (``gunfolds`` graph)
    """
    g = {}
    for subg in glist:
        update_graph(g, subg)
    return g


def ensure_graph_gcd1(g, ignore_singletons=True):
    """
    This function takes any graph, breaks it into SCCs and make sure each SCC has a gcd of 1

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)

    :param ignore_singletons: ignores singleton SCCs when adding a self-loop to make gcd=1
    :type ignore_singletons: boolean

    :returns: a graph with ``gcd=1``
    :rtype: dictionary (``gunfolds`` graph)
    """
    G = graph2nx(g)
    if ignore_singletons:
        x = [ensure_gcd1(subgraph(g, c)) for c in nx.strongly_connected_components(G) if len(c) > 1]
    else:
        x = [ensure_gcd1(subgraph(g, c)) for c in nx.strongly_connected_components(G)]
    return merge_graphs([g] + x)


def gcd1_bp_mean_degree_graph(node_num, degree, seed=None):
    """
    Returns a random graph with ``gcd=1``

    :param node_num: number of nodes
    :type node_num: integer
    
    :param degree: degree
    :type degree: float
    
    :param seed: random seed
    :type seed: integer
    
    :returns: a random graph with ``gcd=1``
    :rtype: dictionary (``gunfolds`` graph)
    """
    g = bp_mean_degree_graph(node_num, degree, seed)
    return ensure_graph_gcd1(g)


def ring_sccs(num, num_sccs, dens=0.5, degree=3, max_cross_connections=3):
    """
    Generate a random graph with ``num_sccs`` SCCs, n-nodes each
    
    :param num: number of nodes in each sec
    :type num: integer
    
    :param num_sccs: number of secs
    :type num_sccs: integer
    
    :param dens: density of each sac
    :type dens: float
    
    :param degree: degree of the connecting DAG
    :type degree: integer
    
    :param max_cross_connections: maximum number of connections per each edge in DAG
    :type max_cross_connections: integer
    
    :returns: a random graph with ``num_sccs`` SCCs, n-nodes each
    :rtype: dictionary (``gunfolds`` graph)
    """
    dag = randomDAG(num_sccs, degree=degree)
    while nx.is_empty(dag):
        dag = randomDAG(num_sccs, degree=degree)
    if not nx.is_weakly_connected(dag):
        randTree = nx.random_tree(n=num_sccs, create_using=nx.DiGraph)
        dag = remove_loop(nx.compose(dag, randTree), dag=dag)
    ss = shift_list_labels([ensure_gcd1(ringmore(num, max(int(dens * num**2)-num, 1))) for i in range(num_sccs)])
    for v in dag:
        v_nodes = [x for x in ss[v].keys()]
        for w in dag[v]:
            w_nodes = [x for x in ss[w].keys()]
            for i in range(np.random.randint(low=1, high=max_cross_connections+1)):
                a = random.choice(v_nodes)
                b = random.choice(w_nodes)
                ss[v][a][b] = 1
    return merge_list(ss)


def selfloop(n, g):
    """
    Checks if ``g`` has a self loop at node n

    :param n: node
    :type n: integer
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :returns: True if ``g`` has a self loop at node n and vice versa
    :rtype: boolean
    """
    return n in g[n]


def remove_loop(G, dag):
    """  
    Removes the loops from ``G`` according to ``dag``

    :param G: graph
    :type G: NetworkX graph
    
    :param dag: DAG
    :type dag: NetworkX graph
    
    :returns: graph with removed loop according to DAG
    :rtype: NetworkX graph
    """
    try:
        while True:
            lis = nx.find_cycle(G)
            for edge in lis:
                if edge in list(dag.edges):
                    G.remove_edge(edge[0], edge[1])
                    break
    except nx.exception.NetworkXNoCycle:
        assert nx.is_directed_acyclic_graph(G)
        assert nx.is_weakly_connected(G)
        return G



from gunfolds.utils import graphkit as gk
from itertools import combinations
import numpy as np
from scipy.spatial.distance import pdist, squareform
from scipy.stats import zscore, norm
import statsmodels.api as sm


def independent(y, X, pval=0.05):
    """
    :param y:
    :type y:

    :param X:
    :type X:

    :param pval:
    :type pval: float

    :returns: 
    :rtype: 
    """
    X = sm.add_constant(X)
    est = sm.OLS(y, X).fit()
    return est.pvalues[1] > pval


def kernel(z):
    """
    :param z:
    :type z:

    :returns: 
    :rtype:
    """
    if np.abs(z) > 1.:
        return 0.
    return .5


def residuals_(x, y, z):
    """
    :param x:
    :type x:
    
    :param y:
    :type y:
    
    :param z:
    :type z:
    
    :returns: 
    :rtype: 
    """
    pd = map(kernel, pdist(z.T))
    PD = squareform(pd)
    sumsx = np.dot(PD, x) + 0.5*x
    sumsy = np.dot(PD, y) + 0.5*y
    weights = np.sum(PD, axis=1) + 0.5
    residualsx = x - sumsx/weights
    residualsy = y - sumsy/weights

    return residualsx, residualsy


def moment22(x, y):
    """
    :param x:
    :type x:
    
    :param y:
    :type y:
    
    :returns: 
    :rtype: 
    """
    return np.dot(x*x, y*y)/len(x)


def fdr(alpha, pvalues):
    """
    :param alpha:
    :type alpha:
    
    :param pvalues:
    :type pvalues:
    
    :returns: 
    :rtype: 
    """
    m = len(pvalues)
    c = np.cumsum(1./(np.arange(m)+1))
    pcompare = (pvalues <= alpha*(np.arange(1, m+1)/(c*(m+1))))
    idx = np.where(pcompare == True)[0]
    if len(idx) == 0:
        return -1
    return idx[-1]

def fdrQ(alpha, pvalues):
    """
    :param alpha:
    :type alpha:
    
    :param pvalues:
    :type pvalues:
    
    :returns: 
    :rtype: 
    """
    pvalues = np.sort(pvalues)
    pvalues = pvalues[~np.isnan(pvalues)]
    min  = np.nan if len(pvalues) == 0 else pvalues[0]
    high = 1.0
    low  = 0.
    q    = alpha
    while (high - low) > 1e-5:
        midpoint = (high + low)/2.0
        q = midpoint
        cutoff = pvalues[fdr(q, pvalues)]

        if cutoff < min:
            low = midpoint
        elif cutoff > min:
            high = midpoint
        else:
            low  = midpoint
            high = midpoint
    return q

def fdrCutoff(alpha, pvalues):
    """
    :param alpha:
    :type alpha:
    
    :param pvalues:
    :type pvalues:
    
    :returns: 
    :rtype: 
    """
    pvalues = np.sort(pvalues)
    k = fdr(alpha, pvalues)
    if k < 0:
        return 0
    return pvalues[k]

def np_fisherZ(x, y, r):
    """
    :param x:
    :type x:
    
    :param y:
    :type y:
    
    :param r:
    :type r:
    
    :returns: 
    :rtype: 
    """
    z = 0.5 * (np.log(1.0 + r) - np.log(1.0 - r))
    w = np.sqrt(len(x)) * z
    x_ = zscore(x)
    y_ = zscore(y)
    t2 = moment22(x_, y_)
    t = np.sqrt(t2)
    p = 2. * (1. - norm.cdf(np.abs(w), 0.0, t))
    return p

def independent_(x, y, alpha = 0.05):
    """
    :param x:
    :type x:
    
    :param y:
    :type y:
    
    :param alpha:
    :type alpha: float
    
    :returns: 
    :rtype: 
    """
    x = x[~np.isnan(x)]
    y = y[~np.isnan(y)]

    # For PC, should not remove the edge for this reason.
    if len(x) < 10:
        return False

    ps = []
    for i in range(15):
        for j in range(15):
            x_ = x**i
            y_ = y**j
            r = np.corrcoef(x_, y_)[0, 1]
            # r = max(min(r,1),-1) # Tetrad had this
            p = np_fisherZ(x_, y_, r)
            # if not np.isnan(p):
            ps.append(p)

    if not ps:
        return True
    return fdrCutoff(alpha, ps) > alpha


def dpc(data, pval=0.05):
    """
    :param data:
    :type data:
    
    :param pval:
    :type pval: float
    
    :returns: 
    :rtype: 
    """
    n = data.shape[0]
    # stack the data: first n rows is t-1 slice, the next n are slice t
    data = np.asarray(np.r_[data[:, :-1], data[:, 1:]])

    def tetrad_cind_(y, x, condset=[], alpha=0.01, shift=0):
        y = data[n+y-1,:]
        x = data[shift+x-1,:]
        if condset:
            X  = data[condset,:]
            ry, rx = residuals_(y, x, X)
        else:
            ry, rx = [y, x]
        return independent_(ry, rx, alpha = alpha)

    def cind_(y,x, condset=[], pval=pval, shift=0):
        yd = data[n+y-1,:].T
        X  = data[[shift+x-1]+condset,:].T
        return independent(yd, X, pval=pval)

    def cindependent(y, x, counter, parents=[], pval=pval):
        for S in [j for j in combinations(parents, counter)]:
            if cind_(y, x, condset=list(S), pval=pval):
            #if tetrad_cind_(x, y, condset=list(S), alpha=pval):
                return True
        return False

    def bindependent(y, x, parents=[], pval=pval):
        return cind_(y, x, condset=parents, pval=pval, shift=n)
        # return tetrad_cind_(y, x, condset=parents, alpha=pval, shift=n)

    def prune(elist, mask, g):
        for e in mask:
            g[e[0]][e[1]] -= 1
            elist.remove(e)
        gk.clean_leaf_nodes(g)

    g  = gk.superclique(n)
    gtr = gk.gtranspose(g)

    el = gk.edgelist(g)
    for counter in range(n):
        to_remove = []
        for e in el:
            ppp = [k-1 for k in gtr[e[1]] if k != e[0]]
            if counter <= len(ppp):
                if cindependent(e[1], e[0], counter, parents=ppp, pval=pval):
                    to_remove.append(e)
                    gtr[e[1]].pop(e[0], None)
        prune(el, to_remove, g)

    bel = [map(lambda k: k+1, x) for x in combinations(range(n), 2)]
    for e in bel:
        ppp = list(set(gtr[e[0]].keys()) | set(gtr[e[1]].keys()))
        ppp = map(lambda x: x-1, ppp)
        if bindependent(e[0], e[1], parents=ppp, pval=pval):
            g[e[0]][e[1]] -= 2
            g[e[1]][e[0]] -= 2
    gk.clean_leaf_nodes(g)

    return g


# Local Variables:
# mode: python
# python-indent-offset: 4
# End:



from gunfolds.conversions import g2num, num2CG, graph2nx
from gunfolds.utils.bfutils import undersample
import networkx
from numpy import argsort


def simple_loops(g, u):
    """
    Iterator over the list of simple loops of graph g at the undersample rate u
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param u: under sampling rate
    :type u: integer
    """
    gx = graph2nx(num2CG(g2num(undersample(g, u)), len(g)))
    for l in networkx.simple_cycles(gx):
        yield l


def print_loops(g, u):
    """
    Prints the simple loops in graph ``g``

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param u: under sampling rate
    :type u: integer
    """
    l = [x for x in simple_loops(g, u)]
    lens = map(len, l)
    for i in argsort(lens):
        print(l[i])



import os
from gunfolds.utils import zickle as zkl

DIR_NAME = os.path.dirname(__file__)
ABS_PATH = os.path.abspath(os.path.join(DIR_NAME))

class LazyLoadData(object):
    """ Prevent import from waiting on pickle file loads """
    @property
    def alloops(self):
        if not hasattr(self, '_alloops'):
            self._alloops = zkl.load('{}/../data/allloops.zkl'.format(ABS_PATH))
        return self._alloops

    @property
    def circp(self):
        if not hasattr(self, '_circp'):
            self._circp = zkl.load('{}/../data/circular_p.zkl'.format(ABS_PATH))
        return self._circp

    @property
    def colors(self):
        if not hasattr(self, '_colors'):
            self._colors = zkl.load('{}/../data/colors.zkl'.format(ABS_PATH))
        return self._colors


load_data = LazyLoadData()



from __future__ import division
import copy
from functools import wraps
from gunfolds.utils import bfutils as bfu
from gunfolds.conversions import g2num, graph2nx
from gunfolds.utils import graphkit as gk
from gunfolds.utils.graphkit import density
import itertools
from networkx import strongly_connected_components
import numpy as np
import random
import scipy
import time
from functools import reduce


def chunks(l, n):
    """ 
    Yield successive n-sized chunks from l.
    
    :param l:
    :type l:
    
    :param n:
    :type n: integer
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def purgepath(path, l):
    """
    :param path:
    :type path:
    
    :param l:
    :type l:
    """
    for i in range(1, len(path) - 1):
        l.remove((path[i], path[i + 1]))


def try_till_d_path(g, d, gt, order=None):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param d:
    :type d:
    
    :param gt:
    :type gt:
    
    :param order:
    :type order:
    
    :returns: 
    :rtype: 
    """
    k = []
    i = 1
    while not k:
        if order:
            k = [x for x in length_d_paths(g, order(i), d)]
        else:
            k = [x for x in length_d_paths(g, i, d)]
        i += 1
        if i > len(g):
            return []

    ld = []
    for i in range(min(10, len(k))):
        ld.append(len(checkApath([2] + k[i], gt)))
    return k[0]


def try_till_path(g, gt):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param gt:
    :type gt:
    
    :returns: 
    :rtype: 
    """
    gx = graph2nx(g)
    sccl = [x for x in strongly_connected_components(gx)]
    # take the largest
    ln = [len(x) for x in sccl]
    idx = np.argsort(ln)
    d = len(sccl[idx[-1]]) - 1
    k = []
    while not k:
        if d < 5:
            return []
        k = try_till_d_path(g, d, gt)
        d -= 1
    return k


def gpurgepath(g, path):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param path:
    :type path:
    """
    for i in range(1, len(path) - 1):
        del g[path[i]][path[i + 1]]


def forks(n, c, el, bl, doempty=lambda x, y: True):
    """
    :param n: single node string
    :type n: (guess)string
    
    :param c: mutable list of children of node n (will be changed as a side effect)
    :type c:
    
    :param el: list of edges available for construction (side effect change)
    :type el:
    
    :param bl: list of bidirected edges
    :type bl:
    
    :param doempty:
    :type doempty:
    
    :returns: 
    :rtype: 
    """
    l = []
    r = set()
    for p in [x for x in itertools.combinations(c, 2)]:
        if doempty(p, bl) and (n, p[0]) in el and (n, p[1]) in el:
            l.append((n,) + p)
            el.remove((n, p[0]))
            el.remove((n, p[1]))
            r.add(p[0])
            r.add(p[1])
    for e in r:
        c.remove(e)
    return l


def childrenedges(n, c, el):
    """
    :param n: single node string
    :type n: (guess)string
    
    :param c: mutable list of children of node n (will be changed as a side effect)
    :type c:
    
    :param el: list of edges available for construction (side effect change)
    :type el:
    
    :returns: 
    :rtype: 
    """
    l = []
    r = set()
    for ch in c:
        if (n, ch) in el:
            l.append((n, ch))
            el.remove((n, ch))
            r.add(ch)
    for e in r:
        c.remove(e)
    return l


def make_emptyforks(n, c, el, bl):
    """ 
    An empty fork is a fork without the bidirected edge 

    :param n: single node string
    :type n: (guess)string
    
    :param c: mutable list of children of node n (will be changed as a side effect)
    :type c:
    
    :param el: list of edges available for construction (side effect change)
    :type el:
    
    :param bl: list of bidirected edges
    :type bl:
    
    :returns: 
    :rtype: 
    """
    return forks(n, c, el, bl, doempty=lambda x, y: not (x in y))


def make_fullforks(n, c, el, bl):
    """
    :param n: single node string
    :type n: (guess)string
    
    :param c: mutable list of children of node n (will be changed as a side effect)
    :type c:
    
    :param el: list of edges available for construction (side effect change)
    :type el:
    
    :param bl: list of bidirected edges
    :type bl:
    
    :returns: 
    :rtype: 
    """
    return forks(n, c, el, bl, doempty=lambda x, y: x in y)


def make_longpaths(g, el):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param el: (GUESS)list of edges available for construction (side effect change)
    :type el:
    
    :returns: 
    :rtype: 
    """
    l = []
    gc = copy.deepcopy(g)
    for i in range(16):
        k = try_till_path(gc, g)
        if len(k) < 5:
            break
        if k:
            l.append((2,) + tuple(k))
            purgepath(l[-1], el)
            gpurgepath(gc, l[-1])
        else:
            break
    return l


def make_allforks_and_rest(g, el, bl, dofullforks=True):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param el: list of edges available for construction (side effect change)
    :type el:
    
    :param bl: (GUESS)list of bidirected edges
    :type bl:
    
    :param dofullforks:
    :type dofullforks:
    
    :returns: 
    :rtype: 
    """
    l = []
    r = []
    nodes = [n for n in g]
    random.shuffle(nodes)
    for n in nodes:

        c = [e for e in g[n] if g[n][e] in (1, 3)]  # all children
        if len(c) == 1:
            if (n, c[0]) in el:
                r.append((n, c[0]))
                el.remove((n, c[0]))
        elif len(c) > 1:
            l.extend(make_emptyforks(n, c, el, bl))
            if dofullforks:
                l.extend(make_fullforks(n, c, el, bl))
            r.extend(childrenedges(n, c, el, bl))
    return l, r


def vedgelist(g, pathtoo=False):
    """ 
    Returns a list of tuples for edges of g and forks
    a superugly organically grown function that badly needs refactoring
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param pathtoo:
    :type pathtoo:
    
    :returns: 
    :rtype: 
    """
    l = []
    el = gk.edgelist(g)
    bl = gk.bedgelist(g)

    if pathtoo:
        l.extend(make_longpaths(g, el))
    l2, r = make_allforks_and_rest(g, el, bl, dofullforks=True)
    l.extend(l2)

    A, singles = makechains(r)

    if singles:
        B, singles = makesinks(singles)
    else:
        B, singles = [], []

    l = longpaths_pick(l) + threedges_pick(l) + A + B + singles
    return l


def twoedges_pick(l):
    """
    :param l:
    :type l:
    
    :returns: 
    :rtype: 
    """
    return [e for e in l if len(e) == 2]


def threedges_pick(l):
    """
    :param l:
    :type l:
    
    :returns: 
    :rtype: 
    """
    return [e for e in l if len(e) == 3]


def longpaths_pick(l):
    """
    :param l:
    :type l:
    
    :returns: 
    :rtype: 
    """
    return [e for e in l if len(e) > 3 and e[0] == 2]


def makechains(l):
    """ 
    Greedily construct 2 edge chains from edge list
    
    :param l:
    :type l:
    
    :returns: 
    :rtype: 
    """
    ends = {e[1]: e for e in l}
    starts = {e[0]: e for e in l}
    r = []
    singles = []
    while l:

        e = l.pop()
        if e[1] in starts and e[0] != e[1] and starts[e[1]] in l:
            r.append((0, e[0],) + starts[e[1]])
            l.remove(starts[e[1]])
        elif e[0] in ends and e[0] != e[1] and ends[e[0]] in l:
            r.append((0,) + ends[e[0]] + (e[1],))
            l.remove(ends[e[0]])
        else:
            singles.append(e)
    return r, singles


def makesink(es):
    """
    :param es:
    :type es:
    
    :returns: 
    :rtype: 
    """
    return (1, es[0][0],) + es[1]


def makesinks(l):
    """ 
    Greedily construct 2 edge sinks ( b->a<-c ) from edge list

    :param l:
    :type l:
    
    :returns: 
    :rtype: 
    """
    sinks = {}
    for e in l:
        if e[1] in sinks:
            sinks[e[1]].append(e)
        else:
            sinks[e[1]] = [e]
    r = []
    singles = []
    for e in sinks:
        if len(sinks[e]) > 1:
            for es in chunks(sinks[e], 2):
                if len(es) == 2:
                    r.append(makesink(es))
                else:
                    singles.append(es[0])
        else:
            singles.append(sinks[e][0])
    return r, singles


def selfloops(l, g):
    """
    :param l:
    :type l:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    return reduce(lambda x, y: x and y, map(lambda x: gk.selfloop(x, g), l))


def checkbedges(v, bel, g2):
    """
    :param v:
    :type v:
    
    :param bel:
    :type bel:
   
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    r = []
    for e in bel:
        if e == tuple(v[1:]) and not selfloops(e, g2):
            r.append(e)
        if e == (v[2], v[1]) and not selfloops(e, g2):
            r.append(e)
    for e in r:
        bel.remove(e)
    return bel


def checkedge(e, g2):
    """
    :param e:
    :type e:
   
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    if e[0] == e[1]:
        l = [n for n in g2 if n in g2[n]]
        s = set()
        for v in g2[e[0]]:
            s.add(g2[e[0]][v])
        if 2 not in s and 3 not in s:
            l.remove(e[0])
        return l
    else:
        return g2.keys()


def single_nodes(v, g2):
    """ 
    Returns a list of singleton nodes allowed for merging with ``v``

    :param v:
    :type v:
   
    :param g2:
    :type g2:
    
    :returns: a list of singleton nodes allowed for merging with ``v``
    :rtype: 
    """
    l = [(n, n) for n in g2 if not (n in v) and len(g2[n]) > 1]
    return l


def checkvedge(v, g2):
    """ 
    Nodes to check to merge the virtual nodes of ``v`` ( b<-a->c )
    
    :param v:
    :type v:
   
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    l = gk.bedgelist(g2)
    if (v[1], v[2]) in l:
        l = single_nodes(v, g2) + checkbedges(v, l, g2)
        for n in v:
            if n in g2[n]:
                l.append((n, n))
    else:
        l = checkbedges(v, l, g2)
    return list(set(l))


def checkAedge(v, g2):
    """ 
    Nodes to check to merge the virtual nodes of A ( b->a<-c )

    :param v:
    :type v:

    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    l = []
    # try all pairs but the sources
    for pair in itertools.combinations(g2, 2):
        # if pair == (v[1],v[2]): continue
        # if pair == (v[2],v[1]): continue
        l.append(pair)
        l.append(pair[::-1])
    for n in g2:
        l.append((n, n))
    return l


def checkcedge(c, g2):
    """ 
    Nodes to check to merge the virtual nodes of ``c`` ( a->b->c )
    
    :param c:
    :type c:
   
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    l = gk.edgelist(g2)
    return list(set(l))


def checkApath(p, g2):
    """
    :param p:
    :type p:
   
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    d = len(p) - 2
    l = []
    for n in g2:
        l.extend([tuple(x) for x in length_d_loopy_paths(g2, n, d, p[1:])])
    # k = prunepaths_1D(g2, p, l)
    return l


def isedge(v):
    """
    :param v:
    :type v:
    
    :returns: 
    :rtype: 
    """
    return len(v) == 2  # a->b


def isvedge(v):
    """
    :param v:
    :type v:
    
    :returns: 
    :rtype: 
    """
    return len(v) == 3  # b<-a->c


def isCedge(v):
    """
    :param v:
    :type v:
    
    :returns: 
    :rtype: 
    """
    return len(v) == 4 and v[0] == 0  # a->b->c


def isAedge(v):
    """
    :param v:
    :type v:
    
    :returns: 
    :rtype: 
    """
    return len(v) == 4 and v[0] == 1  # a->c<-b


def isApath(v):
    """
    :param v:
    :type v:
    
    :returns: 
    :rtype: 
    """
    return len(v) >= 4 and v[0] == 2  # a->b->...->z


def checker(n, ee):
    """
    :param n:
    :type n:
    
    :param ee:
    :type ee:
    
    :returns: 
    :rtype: 
    """
    g = gk.ringmore(n, ee)
    g2 = bfu.increment(g)
    d = checkable(g2)
    t = [len(d[x]) for x in d]
    r = []
    n = len(g2)
    ee = len(gk.edgelist(g2))
    for i in range(1, len(t)):
        r.append(sum(np.log10(t[:i])) - ee * np.log10(n))
    return r


def checkerDS(n, ee):
    """
    :param n:
    :type n:
    
    :param ee:
    :type ee:
    
    :returns: 
    :rtype: 
    """
    g = gk.ringmore(n, ee)
    g2 = bfu.increment(g)
    gg = checkable(g2)
    d, p, idx = conformanceDS(g2, gg, gg.keys())
    t = [len(x) for x in p]
    r = []
    n = len(g2)
    ee = len(gk.edgelist(g2))
    for i in range(1, len(t)):
        r.append(sum(np.log10(t[:i])) - ee * np.log10(n))
    return r


def checkable(g2):
    """
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    d = {}
    g = cloneempty(g2)
    vlist = vedgelist(g2, pathtoo=False)
    for v in vlist:
        if isvedge(v):
            d[v] = checkvedge(v, g2)
        elif isCedge(v):
            d[v] = checkcedge(v, g2)
        elif isAedge(v):
            d[v] = checkAedge(v, g2)
        elif isApath(v):
            d[v] = checkApath(v, g2)
        else:
            d[v] = checkedge(v, g2)

    # check if some of the otherwise permissible nodes still fail
    c = [ok2add2edges,
         ok2addavedge,
         ok2addacedge,
         ok2addaAedge,
         ok2addapath]

    for e in d:
        checks_ok = c[edge_function_idx(e)]
        for n in d[e]:
            if not checks_ok(e, n, g, g2):
                d[e].remove(n)

    return d


def inorder_check2(e1, e2, j1, j2, g2, f=[], c=[]):
    """
    :param e1:
    :type e1:
    
    :param e2:
    :type e2:
    
    :param j1:
    :type j1:
    
    :param j2:
    :type j2:
    
    :param g2:
    :type g2:
    
    :param f:
    :type f:
    
    :param c:
    :type c:
    
    :returns: 
    :rtype: 
    """
    g = cloneempty(g2)  # the graph to be used for checking

    if f == []:
        f = [(add2edges, del2edges, mask2edges),
             (addavedge, delavedge, maskavedge),
             (addacedge, delacedge, maskaCedge),
             (addaAedge, delaAedge, maskaAedge),
             (addapath, delapath, maskapath)]

    if c == []:
        c = [ok2add2edges,
             ok2addavedge,
             ok2addacedge,
             ok2addaAedge,
             ok2addapath]

    adder, remover, masker = f[edge_function_idx(e1)]
    checks_ok = c[edge_function_idx(e2)]

    d = {}
    s1 = set()
    s2 = set()
    for c1 in j1:  # for each connector
        mask = adder(g, e1, c1)
        d[c1] = set()
        for c2 in j2:
            if checks_ok(e2, c2, g, g2):
                d[c1].add(c2)
                s2.add(c2)
        remover(g, e1, c1, mask)
        if d[c1]:
            s1.add(c1)
    return d, s1, s2


def check3(e1, e2, e3, j1, j2, j3, g2, f=[], c=[]):
    """
    :param e1:
    :type e1:
    
    :param e2:
    :type e2:

    :param e3:
    :type e3:
    
    :param j1:
    :type j1:
    
    :param j2:
    :type j2:
    
    :param j3:
    :type j3:
    
    :param g2:
    :type g2:
    
    :param f:
    :type f:
    
    :param c:
    :type c:
    
    :returns: 
    :rtype: 
    """
    g = cloneempty(g2)  # the graph to be used for checking
    if f == []:
        f = [(add2edges, del2edges, mask2edges),
             (addavedge, delavedge, maskavedge),
             (addacedge, delacedge, maskaCedge),
             (addaAedge, delaAedge, maskaAedge),
             (addapath, delapath, maskapath)]
    if c == []:
        c = [ok2add2edges,
             ok2addavedge,
             ok2addacedge,
             ok2addaAedge,
             ok2addapath]

    adder1, remover1, masker1 = f[edge_function_idx(e1)]
    adder2, remover2, masker2 = f[edge_function_idx(e2)]

    checks_ok2 = c[edge_function_idx(e2)]
    checks_ok3 = c[edge_function_idx(e3)]

    s1 = set()
    s2 = set()
    s3 = set()

    for c1 in j1:  # for each connector
        mask1 = adder1(g, e1, c1)
        append_set1 = False
        for c2 in j2:
            append_set2 = False
            if checks_ok2(e2, c2, g, g2):
                mask2 = adder2(g, e2, c2)
                for c3 in j3:
                    if checks_ok3(e3, c3, g, g2):
                        append_set1 = append_set2 = True
                        s3.add(c3)
                remover2(g, e2, c2, mask2)
            if append_set2:
                s2.add(c2)
        if append_set1:
            s1.add(c1)
        remover1(g, e1, c1, mask1)
    return s1, s2, s3


def del_empty(d):
    """
    :param d:
    :type d:
    
    :returns: 
    :rtype: 
    """
    l = [e for e in d]
    for e in l:
        if d[e] == set():
            del d[e]
    return d


def inorder_checks(g2, gg):
    """
    :param g2:
    :type g2:
    
    :param gg:
    :type gg:
    
    :returns: 
    :rtype: 
    """
    # idx = np.argsort([len(gg[x]) for x in gg.keys()])
    # ee = [gg.keys()[i] for i in idx] # to preserve the order
    ee = [e for e in gg]  # to preserve the order
    # cds = conformanceDS(g2, ee)
    # oo = new_order(g2, ee, repeats=100, cds=None)
    # ee = oo[0]
    random.shuffle(ee)
    d = {}  # new datastructure
    d[ee[0]] = {0: gg[ee[0]]}
    for i in range(len(ee) - 1):
        d[ee[i + 1]] = del_empty(inorder_check2(ee[i], ee[i + 1],
                                                gg[ee[i]], gg[ee[i + 1]], g2)[0])
    return ee, d


def cloneempty(g):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    return {n: {} for n in g}  # return a graph with no edges


def ok2addanedge1(s, e, g, g2, rate=1):
    """
    :param s : start
    :type s:
    
    :param e : end
    :type e:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :param rate:
    :type rate:
    
    :returns: 
    :rtype: 
    """
    # directed edges
    # self-loop
    if s == e and not (e in g2[s]):
        return False
    for u in g:  # Pa(s) -> e
        if s in g[u] and not (e in g2[u] and g2[u][e] in (1, 3)):
            return False
    for u in g[e]:  # s -> Ch(e)
        if not (u in g2[s] and g2[s][u] in (1, 3)):
            return False
    # bidirected edges
    for u in g[s]:  # e <-> Ch(s)
        if u != e and not (u in g2[e] and g2[e][u] in (2, 3)):
            return False
    return True


def ok2addanedge2(s, e, g, g2, rate=1):
    """
    :param s : start
    :type s:
    
    :param e : end
    :type e:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
   
    :param g2:
    :type g2:
    
    :param rate:
    :type rate:
    
    :returns: 
    :rtype: 
    """
    mask = addanedge(g, (s, e))
    value = bfu.undersample(g, rate) == g2
    delanedge(g, (s, e), mask)
    return value


def ok2addanedge(s, e, g, g2, rate=1):
    """
    :param s : start
    :type s:
    
    :param e : end
    :type e:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :param rate:
    :type rate:
    
    :returns: 
    :rtype: 
    """
    f = [ok2addanedge1, ok2addanedge2]
    return f[min([1, rate - 1])](s, e, g, g2, rate=rate)


def ok2add2edges(e, p, g, g2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: 
    """
    return edge_increment_ok(e[0], p, e[1], g, g2)


def maskanedge(g, e):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :returns: 
    :rtype: 
    """
    return [e[1] in g[e[0]]]


def mask2edges(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    return [p in g[e[0]], e[1] in g[p]]


def maskavedge(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    return [p[0] in g[e[0]], p[1] in g[e[0]],
            e[1] in g[p[0]], e[2] in g[p[1]]]


def maskaAedge(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    return [p[0] in g[e[1]], p[1] in g[e[2]],
            e[3] in g[p[0]], e[3] in g[p[1]]]


def maskaCedge(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    return [p[0] in g[e[1]], e[2] in g[p[0]],
            p[1] in g[e[2]], e[3] in g[p[1]]]


def maskapath(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    mask = []
    for i in range(len(p)):
        mask.append(p[i] in g[e[i + 1]])
        mask.append(e[i + 2] in g[p[i]])
    return mask


def maskaVpath(g, e, p):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    mask = []
    mask.extend([p[0] in g[e[0]], e[1] in g[p[-1]]])
    for i in range(1, len(p)):
        mask.append(p[i] in g[p[i - 1]])
    return mask


def addanedge(g, e):
    """
    Add edge e[0] -> e[1] to g

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :returns: 
    :rtype: 
    """
    mask = maskanedge(g, e)
    g[e[0]][e[1]] = 1
    return mask


def delanedge(g, e, mask):
    """
    Delete edge e[0] -> e[1] from g if it was not there before

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param mask:
    :type mask:
    
    :returns: 
    :rtype: 
    """
    if not mask[0]:
        g[e[0]].pop(e[1], None)


def add2edges(g, e, p):
    """
    Break edge e[0] -> e[1] into two pieces
    e[0] -> p and p -> e[1]
    and add them to g

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype: 
    """
    mask = mask2edges(g, e, p)
    g[e[0]][p] = g[p][e[1]] = 1
    return mask


def del2edges(g, e, p, mask):
    """
    Restore the graph as it was before adding e[0]->p and p->e[1]

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param mask:
    :type mask:
    
    :returns: 
    :rtype: 
    """
    if not mask[0]:
        g[e[0]].pop(p, None)
    if not mask[1]:
        g[p].pop(e[1], None)


def ok2addavedge(e, p, g, g2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: boolean
    """
    if p[1] == e[0]:
        if p[0] != p[1] and p[0] != e[2] and not (e[2] in g2[p[0]] and g2[p[0]][e[2]] in (2, 3)):
            return False
        if p[0] == p[1] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
            return False
        if p[0] == e[1] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
            return False

    if p[0] == e[0]:
        if p[0] != p[1] and p[1] != e[1] and not (e[1] in g2[p[1]] and g2[p[1]][e[1]] in (2, 3)):
            return False
        if p[0] == p[1] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
            return False
        if p[1] == e[2] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
            return False

    if p[0] == e[1] and p[1] == e[2] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
        return False
    if p[0] == e[2] and not (e[1] in g2[p[1]] and g2[p[1]][e[1]] in (1, 3)):
        return False
    if p[1] == e[1] and not (e[2] in g2[p[0]] and g2[p[0]][e[2]] in (1, 3)):
        return False
    if p[0] == p[1] == e[0] and not (e[2] in g2[e[1]] and g2[e[1]][e[2]] in (2, 3)):
        return False

    if not edge_increment_ok(e[0], p[0], e[1], g, g2):
        return False
    if not edge_increment_ok(e[0], p[1], e[2], g, g2):
        return False

    return True


def addavedge(g, v, b):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    mask = maskavedge(g, v, b)
    g[v[0]][b[0]] = g[v[0]][b[1]] = g[b[0]][v[1]] = g[b[1]][v[2]] = 1
    return mask


def delavedge(g, v, b, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :param mask:
    :type mask:
    """
    if not mask[0]:
        g[v[0]].pop(b[0], None)
    if not mask[1]:
        g[v[0]].pop(b[1], None)
    if not mask[2]:
        g[b[0]].pop(v[1], None)
    if not mask[3]:
        g[b[1]].pop(v[2], None)


def ok2addaAedge(e, p, g, g2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: boolean
    """
    if p[1] == e[1] and not (p[0] in g2[e[2]] and g2[e[2]][p[0]] in (1, 3)):
        return False
    if p[0] == e[2] and not (p[1] in g2[e[1]] and g2[e[1]][p[1]] in (1, 3)):
        return False

    if not edge_increment_ok(e[1], p[0], e[3], g, g2):
        return False
    if not edge_increment_ok(e[2], p[1], e[3], g, g2):
        return False

    return True


def addaAedge(g, v, b):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    mask = maskaAedge(g, v, b)
    g[v[1]][b[0]] = g[v[2]][b[1]] = g[b[0]][v[3]] = g[b[1]][v[3]] = 1
    return mask


def delaAedge(g, v, b, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :param mask:
    :type mask:
    """
    if not mask[0]:
        g[v[1]].pop(b[0], None)
    if not mask[1]:
        g[v[2]].pop(b[1], None)
    if not mask[2]:
        g[b[0]].pop(v[3], None)
    if not mask[3]:
        g[b[1]].pop(v[3], None)


def cleanedges(e, p, g, mask):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param mask:
    :type mask:
    """
    i = 0
    for m in mask:
        if not m[0]:
            g[e[i + 1]].pop(p[i], None)
        if not m[1]:
            g[p[i]].pop(e[i + 2], None)
        i += 1


def cleanVedges(g, e, p, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param mask:
    :type mask:
    """
    if mask:
        if not mask[0]:
            g[e[0]].pop(p[0], None)
        if not mask[1]:
            g[p[-1]].pop(e[1], None)

        i = 0
        for m in mask[2:]:
            if not m:
                g[p[i]].pop(p[i + 1], None)
            i += 1


def ok2addapath(e, p, g, g2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: boolean
    """
    mask = []
    for i in range(len(p)):
        if not edge_increment_ok(e[i + 1], p[i], e[i + 2], g, g2):
            cleanedges(e, p, g, mask)
            return False
        mask.append(add2edges(g, (e[i + 1], e[i + 2]), p[i]))
    cleanedges(e, p, g, mask)
    return True


def ok2addaVpath(e, p, g, g2, rate=2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :param rate:
    :type rate:
    
    :returns: 
    :rtype: boolean
    """
    mask = addaVpath(g, e, p)
    if not gk.isedgesubset(bfu.undersample(g, rate), g2):
        cleanVedges(g, e, p, mask)
        return False
    cleanVedges(g, e, p, mask)
    return True


def addapath(g, v, b):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    mask = maskapath(g, v, b)
    for i in range(len(b)):
        g[v[i + 1]][b[i]] = g[b[i]][v[i + 2]] = 1

    return mask


def addaVpath(g, v, b):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    mask = maskaVpath(g, v, b)
    l = [v[0]] + list(b) + [v[1]]
    for i in range(len(l) - 1):
        g[l[i]][l[i + 1]] = 1
    return mask


def delaVpath(g, v, b, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :param mask:
    :type mask:
    
    :returns: 
    :rtype: 
    """
    cleanVedges(g, v, b, mask)


def delapath(g, v, b, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :param mask:
    :type mask:
    """
    for i in range(len(b)):
        if not mask[2 * i]:
            g[v[i + 1]].pop(b[i], None)
        if not mask[2 * i + 1]:
            g[b[i]].pop(v[i + 2], None)


def prunepaths_1D(g2, path, conn):
    """
    :param g2:
    :type g2:
    
    :param path:
    :type path:
    
    :param conn:
    :type conn:
    
    :returns: 
    :rtype: 
    """
    c = []
    g = cloneempty(g2)
    for p in conn:
        mask = addapath(g, path, p)
        if gk.isedgesubset(bfu.increment(g), g2):
            c.append(tuple(p))
        delapath(g, path, p, mask)
    return c


def ok2addacedge(e, p, g, g2):
    """
    :param e:
    :type e:
    
    :param p:
    :type p:
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype: boolean
    """
    if p[0] == p[1]:
        if not e[2] in g2[e[2]]:
            return False
        if not p[0] in g2[p[0]]:
            return False
        if not (e[3] in g2[e[1]] and g2[e[1]][e[3]] in (1, 3)):
            return False

    if not edge_increment_ok(e[1], p[0], e[2], g, g2):
        return False
    if not edge_increment_ok(e[2], p[1], e[3], g, g2):
        return False

    return True


def addacedge(g, v, b):  # chain
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    mask = maskaCedge(g, v, b)
    g[v[1]][b[0]] = g[v[2]][b[1]] = g[b[0]][v[2]] = g[b[1]][v[3]] = 1
    return mask


def delacedge(g, v, b, mask):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    
    :param b:
    :type b:
    
    :param mask:
    :type mask:
    """
    if not mask[0]:
        g[v[1]].pop(b[0], None)
    if not mask[1]:
        g[b[0]].pop(v[2], None)
    if not mask[2]:
        g[v[2]].pop(b[1], None)
    if not mask[3]:
        g[b[1]].pop(v[3], None)


def esig(l, n):
    """
    Turns edge list into a hash string
    
    :param l:
    :type l:
    
    :param n:
    :type n:
    
    :returns: 
    :rtype: integer
    """
    z = len(str(n))
    n = map(lambda x: ''.join(map(lambda y: str(y).zfill(z), x)), l)
    n.sort()
    n = ''.join(n[::-1])
    return int('1' + n)


def gsig(g):
    """
    Turns input graph ``g`` into a hash string using edges
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype:
    """
    return g2num(g)


def signature(g, edges):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param edges:
    :type edges:
    
    :returns: 
    :rtype:
    """
    return (gsig(g), esig(edges, len(g)))


def supergraphs_in_eq(g, g2, rate=1):
    """
    Find  all supergraphs of ``g``  that are also in  the same equivalence
    class with respect to g2 and the rate.
    Currently works only for bfu.undersample by 1
    
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param g2:
    :type g2:
    
    :param rate:
    :type rate:
    
    :returns: 
    :rtype:
    """
    if bfu.undersample(g, rate) != g2:
        raise ValueError('g is not in equivalence class of g2')

    s = set()

    def addnodes(g, g2, edges):
        if edges:
            masks = []
            for e in edges:
                if ok2addanedge(e[0], e[1], g, g2, rate=rate):
                    masks.append(True)
                else:
                    masks.append(False)
            nedges = [edges[i] for i in range(len(edges)) if masks[i]]
            n = len(nedges)
            if n:
                for i in range(n):
                    mask = addanedge(g, nedges[i])
                    s.add(g2num(g))
                    addnodes(g, g2, nedges[:i] + nedges[i + 1:])
                    delanedge(g, nedges[i], mask)

    edges = gk.edgelist(gk.complement(g))
    addnodes(g, g2, edges)
    return s


def edge_function_idx(edge):
    """
    :param edge:
    :type edge:
    
    :returns: 
    :rtype:
    """
    return min(4, len(edge)) - 2 + min(max(3, len(edge)) - 3, 1) * int(edge[0])


def memo_no_return(func):
    """
    :param func:
    :type func:
    
    :returns: 
    :rtype:
    """
    cache = {}                        # Stored subproblem solutions

    @wraps(func)                      # Make wrap look like func
    def wrap(*args):                  # The memoized wrapper
        s = signature(args[0], args[2])  # Signature: g and edges
        if s not in cache:            # Not already computed?
            cache[s] = func(*args)    # Compute & cache the solution
        return
    return wrap


def v2g22g1(g2, capsize=None, verbose=True):
    """
    Computes all g1 that are in the equivalence class for g2
    
    :param g2:
    :type g2:
    
    :param capsize:
    :type capsize:
    
    :param verbose:
    :type verbose:
    
    :returns: 
    :rtype:
    """
    if bfu.is_sclique(g2):
        print('Superclique - any SCC with GCD = 1 fits')
        return set([-1])

    f = [(add2edges, del2edges, mask2edges),
         (addavedge, delavedge, maskavedge),
         (addacedge, delacedge, maskaCedge),
         (addaAedge, delaAedge, maskaAedge),
         (addapath, delapath, maskapath)]
    c = [ok2add2edges,
         ok2addavedge,
         ok2addacedge,
         ok2addaAedge,
         ok2addapath]

    def predictive_check(g, g2, pool, checks_ok, key):
        s = set()
        for u in pool:
            if not checks_ok(key, u, g, g2):
                continue
            s.add(u)
        return s

    @memo_no_return  # memoize the search
    def nodesearch(g, g2, order, inlist, s, cds, pool, pc):
        if order:
            if bfu.increment(g) == g2:
                s.add(g2num(g))
                if capsize and len(s) > capsize:
                    raise ValueError('Too many elements')
                s.update(supergraphs_in_eq(g, g2))
                return g

            key = order[0]
            if pc:
                tocheck = [x for x in pc if x in cds[len(inlist) - 1][inlist[0]]]
            else:
                tocheck = cds[len(inlist) - 1][inlist[0]]

            if len(order) > 1:
                kk = order[1]
                pc = predictive_check(g, g2, pool[len(inlist)],
                                      c[edge_function_idx(kk)], kk)
            else:
                pc = set()

            adder, remover, masker = f[edge_function_idx(key)]
            checks_ok = c[edge_function_idx(key)]

            for n in tocheck:
                if not checks_ok(key, n, g, g2):
                    continue
                masked = np.prod(masker(g, key, n))
                if masked:
                    nodesearch(g, g2, order[1:], [n] + inlist, s, cds, pool, pc)
                else:
                    mask = adder(g, key, n)
                    nodesearch(g, g2, order[1:], [n] + inlist, s, cds, pool, pc)
                    remover(g, key, n, mask)

        elif bfu.increment(g) == g2:
            s.add(g2num(g))
            if capsize and len(s) > capsize:
                raise ValueError('Too many elements')
            return g

    # find all directed g1's not conflicting with g2

    startTime = time.time()
    gg = checkable(g2)

    idx = np.argsort([len(gg[x]) for x in gg.keys()])
    keys = [gg.keys()[i] for i in idx]

    cds, order, idx = conformanceDS(g2, gg, keys)
    if verbose:
        print("precomputed in {:10.3f} seconds".format(time.time() - startTime))
    if 0 in [len(x) for x in order]:
        return set()
    g = cloneempty(g2)

    s = set()
    try:
        nodesearch(g, g2, [keys[i] for i in idx], [0], s, cds, order, set())
    except ValueError as e:
        print(e)
        s.add(0)
    return s


def conformanceDS(g2, gg, order, f=[], c=[]):
    """
    :param g2:
    :type g2:
    
    :param gg:
    :type gg:
    
    :param order:
    :type order:
    
    :param f:
    :type f:
    
    :param c:
    :type c:
    
    :returns: 
    :rtype:
    """
    CDS = {}
    CDS[0] = set(gg[order[0]])
    pool = [set(gg[order[i]]) for i in range(len(order))]

    for x in itertools.combinations(range(len(order)), 2):

        d, s_i1, s_i2 = inorder_check2(order[x[0]], order[x[1]],
                                       pool[x[0]], pool[x[1]],
                                       g2, f=f, c=c)

        pool[x[0]] = pool[x[0]].intersection(s_i1)
        pool[x[1]] = pool[x[1]].intersection(s_i2)

        d = del_empty(d)
        if not x[1] in CDS:
            CDS[x[1]] = {}
            CDS[x[1]][x[0]] = d
        else:
            CDS[x[1]][x[0]] = d
    if density(g2) > 0.35:
        itr3 = [x for x in itertools.combinations(range(len(order)), 3)]
        for x in random.sample(itr3, min(10, np.int(scipy.special.comb(len(order), 3)))):
            s1, s2, s3 = check3(order[x[0]], order[x[1]], order[x[2]],
                                pool[x[0]], pool[x[1]], pool[x[2]],
                                g2, f=f, c=c)

            pool[x[0]] = pool[x[0]].intersection(s1)
            pool[x[1]] = pool[x[1]].intersection(s2)
            pool[x[2]] = pool[x[2]].intersection(s3)

    return prune_sort_CDS(CDS, pool)


def prune_sort_CDS(cds, pool):
    """
    :param cds:
    :type cds:
    
    :param pool:
    :type pool:
    
    :returns: 
    :rtype:
    """
    idx = np.argsort([len(x) for x in pool])
    p = [pool[i] for i in idx]

    ds = {}
    ds[0] = {}
    ds[0][0] = pool[idx[0]]

    for i in range(1, len(idx)):
        ds[i] = {}
        for j in range(i):
            if idx[j] > idx[i]:
                dd = invertCDSelement(cds[idx[j]][idx[i]])
            else:
                dd = cds[idx[i]][idx[j]]
            for e in pool[idx[i - 1]].intersection(dd.keys()):
                ds[i][e] = pool[idx[i]].intersection(dd[e])

    return ds, p, idx


def invertCDSelement(d_i):
    """
    :param d_i:
    :type d_i:
    
    :returns: 
    :rtype:
    """
    d = {}
    for e in d_i:
        for v in d_i[e]:
            if v in d:
                d[v].add(e)
            else:
                d[v] = set([e])
    return d


def length_d_paths(G, s, d):
    """
    Iterate over nodes in G reachable from s in exactly d steps
    
    :param G:
    :type G:
    
    :param s:
    :type s:
    
    :param d:
    :type d:
    
    :returns: 
    :rtype:
    """
    def recurse(G, s, d, path=[]):
        if d == 0:
            yield path
            return

        for u in G[s]:
            if G[s][u] == 2 or u in path:
                continue
            for v in recurse(G, u, d - 1, path + [u]):
                yield v

    for u in recurse(G, s, d, [s]):
        yield u


def edge_increment_ok(s, m, e, g, g2):
    """
    :param s: start
    :type s:
    
    :param m: middle
    :type m:
    
    :param e : end
    :type e:
    
    :param g:
    :type g:
    
    :param g2:
    :type g2:
    
    :returns: 
    :rtype:
    """
    dd = {1: True, 2: False, 3: True}
    bd = {1: False, 2: True, 3: True}
    # bidirected edges
    for u in g[s]:
        if u != m and not (m in g2[u] and bd[g2[u][m]]):
            return False

    # directed edges
    if s == e:
        if not (m in g2[m] and dd[g2[m][m]]):
            return False
        if not (s in g2[s] and dd[g2[s][s]]):
            return False
    for u in g[m]:
        if not (u in g2[s] and dd[g2[s][u]]):
            return False
        # bidirected edges
        if u != e and not (e in g2[u] and bd[g2[u][e]]):
            return False
    for u in g[e]:
        if not (u in g2[m] and dd[g2[m][u]]):
            return False

    for u in g:
        if s in g[u] and not (m in g2[u] and dd[g2[u][m]]):
            return False
        if m in g[u] and not (e in g2[u] and dd[g2[u][e]]):
            return False

    return True


def length_d_loopy_paths(G, s, dt, p):
    """
    Iterate over nodes in ``G`` reachable from ``s`` in exactly d steps
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param s:
    :type s:
    
    :param dt:
    :type dt:
    
    :param p:
    :type p:
    
    :returns: 
    :rtype:
    """
    g1 = cloneempty(G)

    def recurse(g, g2, s, d, path=[]):

        if edge_increment_ok(p[-d - 2], s, p[-d - 1], g, g2):

            if d == 0:
                yield path
                return

            mask = add2edges(g, (p[-d - 2], p[-d - 1]), s)
            for u in g2[s]:
                if g2[s][u] == 2:
                    continue
                for v in recurse(g, g2, u, d - 1, path + [u]):
                    yield v
            del2edges(g, (p[-d - 2], p[-d - 1]), s, mask)

    for u in recurse(g1, G, s, dt - 1, [s]):
        yield u



from gunfolds.utils import graphkit as gk
from gunfolds import conversions as conv
from itertools import combinations
import numpy as np
import statsmodels.api as sm
import copy

# Written by John Cook
# https://github.com/neuroneural/Undersampled_Graph_Estimation/blob/master/dbnestimation/Tools/grangercausality.py


def gc(data, pval=0.05, bidirected=True):
    """
    :param data: time series data
    :type data: numpy matrix
    
    :param pval: (Ask)
    :type pval: float
    
    :param bidirected: (Ask) 
    :type bidirected: boolean
    
    :returns: 
    :rtype: 
    """
    n = data.shape[0]
    # stack the data: first n rows is t-1 slice, the next n are slice t
    data = np.asarray(np.r_[data[:, :-1], data[:, 1:]])

    def FastPCtest(y, x):
        y = y-1
        x = x-1
        nodelist = list(range(n))
        nodelist.remove(x)
        nodelist.remove(y)
        yd = data[n+int(y), :].T
        #        for L in range(1,len(nodelist)+1):
        #            combs=combinations(nodelist,L)
        #            for sub in combs:
        Xalt = data[:n]
        Xnull = np.vstack([Xalt, data[n+x, :]]).T
        Xalt = Xalt.T
        Xnull = sm.add_constant(Xnull)
        Xalt = sm.add_constant(Xalt)
        estnull = sm.OLS(yd, Xnull).fit()
        estalt = sm.OLS(yd, Xalt).fit()
        diff = sm.stats.anova_lm(estalt, estnull)
        if diff.iloc[1, 5] > pval:
            return True
        return False

    def grangertest(y, x):
        y = data[n+int(y)-1, :].T
        Xnull = data[:n, :].T
        Xalt = np.vstack([data[:(x-1), :], data[x:n, :]]).T
        Xnull = sm.add_constant(Xnull)
        Xalt = sm.add_constant(Xalt)
        estnull = sm.OLS(y, Xnull).fit()
        estalt = sm.OLS(y, Xalt).fit()
        diff = sm.stats.anova_lm(estalt, estnull)
        return diff.iloc[1, 5] > pval
    
    new_g = gk.superclique(n)
    # print(new_g)
    g = conv.ian2g(new_g)
    for i in range(1, n+1):
        for j in range(1, n+1):
            if grangertest(j, i):
                sett = copy.deepcopy(g[str(i)][str(j)])
                sett.remove((0, 1))
                g[str(i)][str(j)] = sett
    biedges = combinations(range(1, n+1), 2)
    for i, j in biedges:
        if bidirected:
            if FastPCtest(j, i):
                sett = copy.deepcopy(g[str(i)][str(j)])
                sett.remove((2, 0))
                g[str(i)][str(j)] = sett
                g[str(j)][str(i)] = sett
        else:
            sett = copy.deepcopy(g[str(i)][str(j)])
            sett.remove((2, 0))
            g[str(i)][str(j)] = sett
            g[str(j)][str(i)] = sett
    # print(g)
    return conv.dict_format_converter(g)



from setuptools import find_packages, setup

with open('README.md') as description:
    description = description.read()

exec(open('version.py').read()) # Acquire version constants.

# Define some package entry points. These will be command-line scripts that get
# installed into the user's PATH
epoints = """
[console_scripts]
"""

install_requires = ['networkx', 'igraph', 'scipy', 'numpy',
                    'statsmodels', 'sympy', 'seaborn', 'pandas',
                    'matplotlib', 'progressbar', 'ipdb', 'pydot', 'clingo']

tests_require = install_requires

setup(
    name='gunfolds',
    description='Tools to explore dynamic causal graphs in the case of undersampled data',
    version=__version__,
    long_description_content_type="text/markdown",
    author='Sergey Plis, Cynthia Freeman, Ian Beaver',
    author_email='splis@mrn.org',
    license='GPL',
    long_description=description,
    include_package_data=True, # Include files listed in MANIFEST.in
    packages=find_packages(),
    #entry_points=epoints,
    install_requires=install_requires, # List of dependencies.
    test_suite='tests.tests',
    tests_require=tests_require,
    zip_safe=False) # Override annoying default behavior of easy_install.
